from .attributes import (Attribute, 
                         FeatureAttributes, FeatureAttributesLog, FeatureAttributeBulkCreate,
                         CollectionAttributes, CollectionAttributesLog, CollectionAttributeBulkCreate)
from .column import Column, ColumnCreate, ColumnUpdate
from .data_source import DataSource, DataSourceCreate, DataSourceUpdate, DataSourceColumn
from .dataframe import Dataframe, DataframeCreate, DataframeUpdate
from .dimensionality import Dimensionality, DimensionalityCreate
from .enums import DataType
from .feature import FeatureCreate, FeatureUpdate, FeatureStats
from .feature_set import FeatureSet, FeatureSetCreate, FeatureSetUpdate, FeatureSetYML, BasicFeatureSet
from .organization import Organization



from typing import Dict, List, Optional, Union

import pandas as pd
import time
from tqdm import tqdm

from .feature import Feature, FeatureList
from pyrasgo.api.connection import Connection
from pyrasgo.schemas.attributes import Attribute, CollectionAttributes, CollectionAttributeBulkCreate
from pyrasgo.schemas.enums import ModelType
from pyrasgo.schemas import collection as schema
from pyrasgo.utils import track_usage

class Collection(Connection):
    """
    Stores a Rasgo Collection
    """

    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.Collection(**api_object)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Collection(id={self.id}, name={self.name}, type={self.modelType.name}, authorId={self.authorId}, isShared={self.isShared})"

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self.refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

# ----------
# Properties
# ----------
    @property
    def attributes(self) -> dict:
        """
        Helper property to convert a list of dicts into a single dict
        """
        attr_dict = {}
        response = self._get(f"/models/{self.id}/attributes", api_version=1).json()
        for a in response:
            attr_dict.update({a['key']: a['value']})
        return attr_dict

    @property
    def author(self) -> str:
        """
        Returns the author object of the Collection (includes credentials if user is author)
        """
        if self._profile['id'] == self.authorId:
            author = self._profile
        else:
            author = self._get(f"/users/{self.authorId}", api_version=1).json()
        return f'Author({author["id"]}, {author["firstName"]} {author["lastName"]})'

    @property
    def dimensions(self):
        return self._fields.dimensions

    @property
    def features(self) -> FeatureList:
        """
        Returns the Features within the Collection
        """
        try:
            return FeatureList([feature.dict() for feature in self._fields.features])
        except AttributeError:
            self.refresh()
        return FeatureList([feature.dict() for feature in self._fields.features])

    @features.setter
    def features(self, features: Union[List[Feature], FeatureList, Feature]):
        # Update the namespace
        self._fields.__setattr__('features',
                                    [Feature(api_object=feature.dict()) for feature in self.features + features])

    @property
    def modelType(self) -> ModelType:
        """
        Returns the type of the Collection
        """
        return ModelType(self._fields.type)

# -------
# Methods
# -------
    @track_usage
    def add_attributes(self, attributes: List[dict]) -> List[dict]:
        if not isinstance(attributes, list):
            raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
        attr = []
        for kv in attributes:
            if not isinstance(kv, dict):
                raise ValueError('attributes parameter must be passed in as a list of k:v pairs. Example: [{"key": "value"}, {"key": "value"}]')
            for k, v in kv.items():
                attr.append(Attribute(key=k, value=v))
        attr_in = CollectionAttributeBulkCreate(collectionId = self.id, attributes=attr)
        return self._put(f"/models/{self.id}/attributes", attr_in.dict(exclude_unset=True), api_version=1).json()

    @track_usage
    def add_feature(self, feature: Feature) -> None:
        """
        Adds a single feature to the Collection
        """
        self.features = feature
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id)]})

    @track_usage
    def add_features(self, features: Union[List[Feature], FeatureList]) -> None:
        """
        Adds a FeatureList or a list of Features to the Collection
        """
        self.features = features
        self._patch(f"/models/{self.id}/features", api_version=1,
                    _json={"featureIds": [str(feature.id) for feature in features]})

    @track_usage
    def generate_training_data(self, trigger_stats: bool = True) -> None:
        """
        Triggers the generation of the Collection's training data.
        """
        self._post(f"/trainings/",
                   _json={"modelId": int(self.id),
                          "userId": self._profile['id']},
                   params={"trigger_stats": trigger_stats} if trigger_stats is not None else {},
                   api_version=0)
        for _ in tqdm(range(0, 10), leave=False):
            if self.is_data_ready():
                break
            time.sleep(1)

    @track_usage
    def is_data_ready(self) -> bool:
        """
        Performs check against API for training data readiness, if true, a dataframe can be pulled down.
        :return:
        """
        r = self._get("/trainings/latest", api_version=0)
        if any([model['state'] == 'done' for model in r.json() if model['model']['id'] == self.id]):
            return True
        if any([model['state'] == 'error' for model in r.json() if model['model']['id'] == self.id]):
            raise SystemError("There's been an issue with the training data generation")
        return False

    @track_usage
    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        from pyrasgo.api.read import Read
        return Read().collection_data(id=self.id, filters=filters, limit=limit)

    @track_usage
    def refresh(self) -> None:
        """
        Updates the Collection's attributes from the API
        """
        self._fields = Feature(api_object=self._get(f"/models/{self.id}", api_version=1).json())

    @track_usage
    def rename(self, new_name: str) -> None:
        """
        Updates a Collection's display name
        """
        print(f"Renaming Collection {self.id} from {self.name} to {new_name}")
        collection = schema.CollectionUpdate(id=self.id, name=new_name)
        self._fields = schema.Collection(**self._patch(f"/models/{self.id}/details",
                                                    api_version=1, _json=collection.dict(exclude_unset=True, exclude_none=True)).json())

    @track_usage
    def render_sql_definition(self) -> str:
        sql_str = self._get(f"/models/{self.id}/sql", api_version=1).json()
        return sql_str

    @track_usage
    def share(self, share: bool=True) -> bool:
        """
        Share the Collection with your organization
        :param share: True to make your Collection public. False to make your Collection private
        """
        response = self._patch(f"/models/{self.id}/details", api_version=1, _json={"isShared": share}).json()
        return response['isShared']

    @track_usage
    def toggle_stats(self, trigger_stats: bool) -> dict:
        """
        Determine whether to run correlation and join stats when training the Collection
        """
        response = self.add_attributes([{"_trigger_stats": trigger_stats}])
        for r in response:
            if r.get("key") == "_trigger_stats":
                return {r["key"]: r["value"]}

    def _make_table_metadata(self) -> dict:
        organization = self._get_profile().get("organization")
        metadata = {
            "database": organization.get("database"),
            "schema": organization.get("schema"),
            "table": self._fields.dataTableName,
        }
        return metadata


from setuptools import setup
import os

_here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(_here, 'pyrasgo', 'version.py')) as f:
    exec(f.read(), version)

with open(os.path.join(_here, 'DESCRIPTION.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyrasgo',
    version=version['__version__'],
    description=('Alpha version of the Rasgo Python interface.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Patrick Dougherty',
    author_email='patrick@rasgoml.com',
    url='https://www.rasgoml.com/',
    license='MPL 2.0',
    packages=['pyrasgo',
              'pyrasgo.api',
              'pyrasgo.primitives',
              'pyrasgo.schemas',
              'pyrasgo.storage',
              'pyrasgo.storage.dataframe',
              'pyrasgo.storage.datawarehouse',
              'pyrasgo.utils'],
    install_requires=[
        # Note these are duplicated in requirements.txt
        "more-itertools",
        "pandas",
        "pydantic",
        "pyyaml",
        "requests",
        "tqdm",
        "snowflake-connector-python>=2.4.0",
        "snowflake-connector-python[pandas]",
        "idna>=2.5,<3",
        "pyarrow>=3.0.0,<4"
    ],
    extras_require={
        "df":  [
            "shap", 
            "catboost"
        ],
        "snowflake":  [
            "snowflake-connector-python>=2.4.0",
            "snowflake-connector-python[pandas]",
            "idna>=2.5,<3",
            "pyarrow>=3.0.0,<4"
        ],
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7'],
)



from pydantic import BaseModel, Field
from typing import Any, List, Optional

from pyrasgo.schemas.attributes import Attribute
from pyrasgo.schemas.data_source import DataSource
from pyrasgo.schemas.enums import DataType
from pyrasgo.schemas.granularity import Granularity


# Schemas to be used for general pyrasgo
class Dimension(BaseModel):
    columnName: str
    dataType: DataType
    granularity: Optional[str]

class Feature(BaseModel):
    id: Optional[int]
    columnName: str
    dataType: DataType
    description: Optional[str]
    displayName: Optional[str]
    status: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]
    class Config:
        allow_population_by_field_name = True

class FeatureSet(BaseModel):
    id: Optional[int]
    name: Optional[str]
    sourceTable: str
    sourceCode: Optional[str]
    dimensions: Optional[List[Dimension]]
    features: Optional[List[Feature]]
    dataSource: Optional[DataSource]
    granularities: Optional[List[Granularity]]


class BasicFeatureSet(BaseModel):
    id: Optional[int]
    name: Optional[str]
    snowflakeTable: str = Field(alias='sourceTable')
    sourceCode: Optional[str]
    dataSourceId: Optional[int]
    class Config:
        allow_population_by_field_name = True


class FeatureSetCreate(BaseModel):
    name: str
    snowflakeTable: str
    dataSourceId: int
    rawFilePath: Optional[str]
    sourceCode: Optional[str]


class FeatureSetUpdate(BaseModel):
    id: Optional[int]
    name: Optional[str]
    sourceCode: Optional[str]


# Schema to be used for YML files
# TODO: sync with new schema when customers' legacy files are converted
class DimensionYML(BaseModel):
    columnName: str
    dataType: str
    granularity: str
    class Config:
        allow_population_by_field_name = True

class FeatureYML(BaseModel):
    columnName: str
    dataType: str
    description: Optional[str]
    displayName: Optional[str]
    tags: Optional[List[str]]
    class Config:
        allow_population_by_field_name = True

class DataSourceYML(BaseModel):
    name: Optional[str]

class FeatureSetYML(BaseModel):
    name: Optional[str] = ''
    sourceTable: str
    dimensions: Optional[List[DimensionYML]]
    features: Optional[List[FeatureYML]]
    dataSource: Optional[DataSourceYML]
    script: Optional[str] = ''
    sourceCode: Optional[str]
    class Config:
        allow_population_by_field_name = True